use std::{ffi::OsStr, path::PathBuf};

use tokio::fs;
use tracing::{debug, error, instrument, trace, warn};
use url::Url;

#[instrument(skip(data))]
/// Returns whether or not the saved file should be parsed.
/// If the file is just data, like an image, it doesn't need to be parsed.
/// If it's html, then it does need to be parsed.
pub async fn store(data: &str, url: &Url) -> bool {
    // extract data from url to save it accurately
    let url_path = PathBuf::from("./downloaded/".to_string() + url.domain().unwrap_or("UnknownDomain") + url.path());

    // if it's a file
    let (basepath, filename) = if url_path.extension().filter(valid_file_extension).is_some() {
        // get everything up till the file
        let basepath = url_path.ancestors().skip(1).take(1).collect::<PathBuf>();
        // get the file name
        let filename = url_path.file_name().expect("This should exist").to_string_lossy();
        trace!("Save path: {:?} and base path: {:?}", &url_path, &basepath);
        (basepath, filename.to_string())
    } else {
        (url_path.clone(), "index.html".into())
    };

    let should_parse = filename.ends_with(".html");

    debug!("Writing at: {:?} {:?}", basepath, filename);

    // create the folders
    if let Err(err) = fs::create_dir_all(&basepath).await {
        error!("Dir creation: {err} {:?}", basepath);
    } else {
        if let Err(err) = fs::write(&basepath.join(filename), data).await {
            error!("File creation: {err} {:?}", url_path);
        }
    }

    should_parse
}

fn valid_file_extension(take: &&OsStr) -> bool {
    let los = take.to_string_lossy();
    let all = los.split('.');
    match all.last() {
        Some(s) => {
            // FIXME it's worth noting that the dumb tlds like .zip are in here,
            // which could cause problems
            let all_domains = include_str!("tlds-alpha-by-domain.txt");

            // check if it is a domain
            match all_domains.lines().map(str::to_lowercase).find(|x| x==s.to_lowercase().as_str()) {
                Some(_) => false,
                None => true
            }
        },
        None => false,
    }
}
